/*****************************************************************************
 * Copyright (c) 2020, rrcc FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include "Assembler.h"
#include "instructions/End.h"

void Assembler::assemble(uint32_t baseAddress, uint32_t imageSize) {
    entries.clear();
    bytes.clear();
    bytes.resize(imageSize - 1, 0x00);
    //bytes.resize(imageSize - 1, (uint8_t) Instruction::Instruction::CommandOp::END);

    uint32_t freeOffsetLocation = baseAddress;

    if(linker.getVariables().find("rrcc_ENTRYPOINT") == linker.getVariables().end()){
        throw std::runtime_error("rrcc_ENTRYPOINT is not defined");
    }

    std::string entrypointName;

    for(auto& t : linker.getVariables().at("rrcc_ENTRYPOINT")){
        if(t.getType() == Token::Type::CodeLabel){
            entrypointName = t.getTextValue();
            break;
        }
    }

    if(entrypointName.empty()){
        throw std::runtime_error("rrcc_ENTRYPOINT is not set");
    }


    if (placementMode == FunctionPlacementMode::PlaceAll) {
        for (auto &f : linker.getTree()) {
            if (f.label == entrypointName) {
                baseAddress = freeOffsetLocation;
            }
            freeOffsetLocation = placeFunction(freeOffsetLocation, f);
        }
    } else {
        auto entrypoint = getFunctionByLabel(entrypointName);
        if (entrypoint == nullptr) {
            throw std::runtime_error("Entrypoint function "+entrypointName+" not found");
        }

        freeOffsetLocation = placeFunction(freeOffsetLocation, *entrypoint);
    }

    while (true) {
        bool changed = false;
        for (auto &fEntry : entries) {
            if (fEntry.second.second) {
                auto f = getFunctionByLabel(fEntry.first);
                for (auto &d : f->declarations) {
                    for (auto &t : d.tokens) {
                        if (t.getType() == Token::Type::CodeLabel && t.getTextValue().at(0) != '.') {
                            if (entries.find(t.getTextValue()) == entries.end()) {
                                changed = true;
                                auto fUsed = getFunctionByLabel(t.getTextValue());
                                if (fUsed == nullptr) {
                                    throw std::runtime_error(
                                            t.getTextValue() + " function not found, in use by " + fEntry.first);
                                }

                                freeOffsetLocation = placeFunction(freeOffsetLocation, *fUsed);
                            }
                        }
                    }
                }

                placeFunction(fEntry.second.first, *f);
            }
        }

        if (!changed) {
            break;
        }
    }

    setSignature("Image generated with rrc-as.");


    bytes[0] = 0x38;
    bytes[1] = (baseAddress >> 16) & 0xFF;
    bytes[2] = (baseAddress >> 8) & 0xFF;
    bytes[3] = baseAddress & 0xFF;
}

uint32_t Assembler::placeFunction(uint32_t base, const Function &function) {
    std::unordered_map<std::string, uint32_t> labeledAddress;
    std::unordered_map<uint32_t, uint32_t> endAddresses;

    auto declarations = function.declarations; //Copy is necessary

    uint32_t offset = base;

    //Resolve local labels and end addresses
    for (auto &d : declarations) {
        if (!d.label.empty()) {
            labeledAddress[d.label] = offset;
        }

        bool hasOperator = false;
        for (auto &t : d.tokens) {
            if (t.getType() == Token::Type::Operator) {
                hasOperator = true;
                break;
            }
        }

        if (!hasOperator) {
            continue;
        }

        auto instr = Instruction::Instruction::getCommandForDeclaration(d);
        if (instr == nullptr) {
            throw std::runtime_error("Unknown or invalid declaration");
        }

        instr->setAddress(offset);

        offset += instr->getByteSize();

        endAddresses[instr->getAddress()] = offset;
    }

    offset = base;

    bool needsReLabeling = false;

    for (auto &d : declarations) {
        for (auto &t : d.tokens) {
            if (t.getType() == Token::Type::CodeLabel) {
                if (t.getTextValue().at(0) == '.') {
                    if (labeledAddress.find(t.getTextValue()) != labeledAddress.end()) {
                        t.valueImmediate += labeledAddress.at(t.getTextValue());
                    } else {
                        throw std::runtime_error("Could not find local label " + t.getTextValue());
                    }
                } else if (entries.find(t.getTextValue()) != entries.end()) {
                    t.valueImmediate += entries.at(t.getTextValue()).first;
                } else {
                    needsReLabeling = true;
                }
            }
        }
    }

    for (auto &d : declarations) {
        auto instr = Instruction::Instruction::getCommandForDeclaration(d);
        if (instr == nullptr) {
            throw std::runtime_error("Unknown or invalid declaration");
        }

        if (!d.label.empty()) {
            labeledAddress[d.label] = offset;
        }

        instr->setAddress(offset);

        instr->setEndAddress(endAddresses[offset]); //TODO: this might not be needed anymore!

        auto data = instr->toBytes();

        if (offset + data.size() > bytes.size()) {
            throw std::runtime_error("Instruction encoding exceeds image size");
        }

        std::copy(data.begin(), data.end(), bytes.begin() + offset);
        offset += data.size();
    }

    if (padEndFunction) {
        auto endInstruction = Instruction::End();
        endInstruction.reserved = base; // Tag start of function
        auto data = endInstruction.toBytes();

        std::copy(data.begin(), data.end(), bytes.begin() + offset);
        offset += data.size();
    }

    entries[function.label] = {base, needsReLabeling};

    return offset; //Free bytes between functions
}

const Function *Assembler::getFunctionByLabel(const std::string &l) {
    for (auto &f : linker.getTree()) {
        if (f.label == l) {
            return &f;
        }
    }

    return nullptr;
}

void Assembler::setSignature(const std::string &str) {
    const uint32_t signatureOffset = 0x100;
    std::copy(str.begin(), str.end(), bytes.begin() + signatureOffset);
    bytes[signatureOffset + str.size()] = 0xFF;
}
